from pathlib import Path
from typing import Dict, Text
from unittest.mock import Mock

import pytest
from _pytest.monkeypatch import MonkeyPatch

from rasax.community.services.git_service import GitService
from sanic.testing import SanicTestClient
from sqlalchemy.orm import Session

REPOSITORIES_URL = "/api/projects/default/git_repositories"

TEST_REPOSITORY = {
    "name": "my-repo",
    "repository_url": "my-url",
    "ssh_key": "safe-key",
    "git_service": "Github",
    "git_service_access_token": "token",
    "target_branch": "master",
}


@pytest.fixture()
def test_repository(session: Session, monkeypatch: MonkeyPatch, tmpdir: Path) -> Dict:
    from tests.unit.services.test_git_service import RASA_REMOTE_REPOSITORY

    test_repo = TEST_REPOSITORY.copy()
    test_repo.update(RASA_REMOTE_REPOSITORY)

    return test_repo


async def test_add_repository(
    test_client: SanicTestClient, auth_header: Dict[Text, Text], test_repository: Dict
):
    # create repository
    response = await test_client.post(
        REPOSITORIES_URL, json=test_repository, headers=auth_header
    )

    assert response.status == 201

    # Check response body
    response_body = await response.json()
    assert response_body.pop("id")

    fields_to_check = test_repository.keys() - {"ssh_key", "git_service_access_token"}
    assert all(
        [response_body[field] == test_repository[field] for field in fields_to_check]
    )


async def test_add_repository_with_invalid_body(
    test_client: SanicTestClient, auth_header: Dict[Text, Text]
):
    invalid_repository = {"name": "test"}

    response = await test_client.post(
        REPOSITORIES_URL, json=invalid_repository, headers=auth_header
    )

    assert response.status == 400


async def test_add_repository_with_in_sufficient_permissions(
    test_client: SanicTestClient,
    auth_header: Dict[Text, Text],
    monkeypatch: MonkeyPatch,
):
    invalid_repository = {"repository_url": "test"}

    save_repository_mock = Mock(side_effect=ValueError())
    monkeypatch.setattr(GitService, "save_repository", save_repository_mock)

    response = await test_client.post(
        REPOSITORIES_URL, json=invalid_repository, headers=auth_header
    )

    assert response.status == 422


async def test_list_repositories(
    test_client: SanicTestClient, auth_header: Dict[Text, Text], test_repository: Dict
):
    test_repository = test_repository.copy()
    different_name = "test-list-repositories"
    test_repository["name"] = different_name
    # Create repository
    _ = await test_client.post(
        REPOSITORIES_URL, json=test_repository, headers=auth_header
    )

    # List repositories
    response = await test_client.get(REPOSITORIES_URL, headers=auth_header)

    assert response.status == 200
    repositories = await response.json()
    assert different_name in [repository["name"] for repository in repositories]


async def test_get_single_repository(
    test_client: SanicTestClient, auth_header: Dict[Text, Text], test_repository: Dict
):
    # Create repository
    response = await test_client.post(
        REPOSITORIES_URL, json=test_repository, headers=auth_header
    )
    id_of_created = (await response.json())["id"]

    # Get repository
    response = await test_client.get(
        f"{REPOSITORIES_URL}/{id_of_created}", headers=auth_header
    )
    response_body = await response.json()

    fields_to_check = test_repository.keys() - {"ssh_key", "git_service_access_token"}
    assert all(
        [response_body[field] == test_repository[field] for field in fields_to_check]
    )


async def test_update_repository(
    test_client: SanicTestClient, auth_header: Dict[Text, Text], test_repository: Dict
):
    # Create repository
    response = await test_client.post(
        REPOSITORIES_URL, json=test_repository, headers=auth_header
    )
    id_of_created = (await response.json())["id"]

    updated = test_repository.copy()
    updated_name = "updated 🎉"
    updated["name"] = updated_name

    # Update repository
    response = await test_client.put(
        f"{REPOSITORIES_URL}/{id_of_created}", json=updated, headers=auth_header
    )

    assert response.status == 204

    # Get repository
    response = await test_client.get(
        f"{REPOSITORIES_URL}/{id_of_created}", headers=auth_header
    )
    response_body = await response.json()
    assert response_body["name"] == updated_name


async def test_update_repository_with_invalid_id(
    test_client: SanicTestClient, auth_header: Dict[Text, Text], test_repository: Dict
):
    # Update repository
    response = await test_client.put(
        f"{REPOSITORIES_URL}/56789", json=test_repository, headers=auth_header
    )

    assert response.status == 404


async def test_delete_repository(
    test_client: SanicTestClient, auth_header: Dict[Text, Text], test_repository: Dict
):
    # Create repository
    response = await test_client.post(
        REPOSITORIES_URL, json=test_repository, headers=auth_header
    )
    id_of_created = (await response.json())["id"]

    # Delete repository
    response = await test_client.delete(
        f"{REPOSITORIES_URL}/{id_of_created}", headers=auth_header
    )

    assert response.status == 204

    # Check repository is gone
    response = await test_client.get(
        f"{REPOSITORIES_URL}/{id_of_created}", headers=auth_header
    )

    assert response.status == 404


async def test_delete_repository_with_invalid_id(
    test_client: SanicTestClient, auth_header: Dict[Text, Text]
):
    # Delete repository
    response = await test_client.delete(
        f"{REPOSITORIES_URL}/456789", headers=auth_header
    )

    assert response.status == 404


async def test_get_git_status(
    test_client: SanicTestClient,
    auth_header: Dict[Text, Text],
    test_repository: Dict,
    session: Session,
    monkeypatch: MonkeyPatch,
):
    expected = {
        "is_committing_to_target_branch_allowed": False,
        "is_remote_ahead": False,
        "are_there_local_changes": False,
        "current_branch": "master",
        "time_of_last_pull": 123,
    }
    monkeypatch.setattr(GitService, "get_repository_status", lambda _: expected)

    response = await test_client.get(
        f"{REPOSITORIES_URL}/1/status", headers=auth_header
    )

    assert response.status == 200

    response_body = await response.json()
    assert response_body == expected


@pytest.mark.parametrize("force", [True, False])
async def test_change_branch(
    test_client: SanicTestClient,
    auth_header: Dict[Text, Text],
    monkeypatch: MonkeyPatch,
    force: bool,
):
    mock = Mock()
    monkeypatch.setattr(GitService, "checkout_branch", mock)

    # Checkout other branch
    other_branch = "so-different"
    response = await test_client.put(
        f"{REPOSITORIES_URL}/1/branches/{other_branch}?force={force}",
        headers=auth_header,
    )

    assert response.status == 204
    mock.assert_called_with(other_branch, force=force)


async def test_create_commit(
    test_client: SanicTestClient,
    auth_header: Dict[Text, Text],
    monkeypatch: MonkeyPatch,
    test_repository: Dict,
):
    commit_object = {"test": "this was patched"}

    commit_mock = Mock(return_value=commit_object)
    monkeypatch.setattr(GitService, "commit_and_push_changes_to", commit_mock)

    # Commit changes (stage, commit, push)
    branch = "master"
    response = await test_client.post(
        f"{REPOSITORIES_URL}/1/branches/{branch}/commits",
        json=test_repository,
        headers=auth_header,
    )

    assert response.status == 201
    response_body = await response.json()
    assert response_body == commit_object

    commit_mock.assert_called_with(branch)
    commit_mock.assert_called_once()
